import { callUnomi } from "../utils/connection";
import { validateSegmentObject, reformatSegment } from "../utils/validation";
import { FilteredResponse } from "../../src/types/sdkResponse";
import { UsedProperties } from "../../src/types/segments";

async function resolveReformatSegment(url: string, headers: Record<string, string>): FilteredResponse { // format segment to have correct form
    const callObjectPromise = await callUnomi("GET", url, null, headers, 200); // get segment

    if (callObjectPromise.statusCode === 200) { // if call success
        return new Promise((resolve, reject) => { // return formatted segment result object as a new promise
            resolve(reformatSegment(callObjectPromise, callObjectPromise.responseData)); // format segment and return object that needs to be resolved
        });
    }

    else { // call not success
        return new Promise((resolve, reject) => { // return result object as a new promise
            resolve(callObjectPromise); // object that needs to be resolved
        });
    }
}

/**
 * @function getAll
 * @param {string} baseUrl 
 * @param {Record<string, string>} headers 
 * @param {string} sortOrder 
 * @param {number} limit 
 * @param {number} offset 
 * @returns {FilteredResponse}
 */
export function getAll(baseUrl: string, headers: Record<string, string>, sortOrder: string, limit: number, offset: number): FilteredResponse { // get all segments
    const url = baseUrl + `/cxs/segments?size=${limit}&offset=${offset}&sort=${sortOrder}`; // segment endpoint

    return callUnomi("GET", url, null, headers, 200); // return result object
}

/**
 * @function get
 * @param {string} baseUrl 
 * @param {Record<string, string>} headers 
 * @param {string} segmentId 
 * @returns {FilteredResponse}
 */
export function get(baseUrl: string, headers: Record<string, string>, segmentId: string): FilteredResponse { // get specific segment
    const url = baseUrl + `/cxs/segments/${segmentId}`; // segment endpoint

    return resolveReformatSegment(url, headers); // return formatted segment
}

/**
 * @function create
 * @param {string} baseUrl 
 * @param {Record<string, string>} headers 
 * @param {object} properties 
 * @returns {FilteredResponse}
 */
export function create(baseUrl: string, headers: Record<string, string>, params: UsedProperties): FilteredResponse { // create segment
    const url = baseUrl + `/cxs/segments`; // segment endpoint

    const validatedSegmentObject = validateSegmentObject(params, false); // validate given segment

    if (validatedSegmentObject.areSubConditionsValid) { // check if subconditions are valid
        return callUnomi("POST", url, validatedSegmentObject.segmentObject, headers, 204); // return result object
    }

    else { // subconditions are not valid
        return callUnomi("POST", url, validatedSegmentObject.segmentObject, headers, 204, validatedSegmentObject.errors); // return result object
    }
}

/**
 * @function update
 * @param {string} baseUrl 
 * @param {Record<string, string>} headers 
 * @param {object} properties 
 * @returns {FilteredResponse}
 */
export function update(baseUrl: string, headers: Record<string, string>, params: UsedProperties): FilteredResponse { // update segment
    const url = baseUrl + `/cxs/segments`; // segment endpoint

    const validatedSegmentObject = validateSegmentObject(params, true); // validate given segment

    if (validatedSegmentObject.areSubConditionsValid) { // check if subconditions are valid
        return callUnomi("POST", url, validatedSegmentObject.segmentObject, headers, 204); // return result object
    }

    else { // subconditions are not valid
        return callUnomi("POST", url, validatedSegmentObject.segmentObject, headers, 204, validatedSegmentObject.errors); // return result object
    }
}

/**
 * @function delete
 * @param {string} baseUrl 
 * @param {Record<string, string>} headers 
 * @param {string} segmentId 
 * @returns {FilteredResponse}
 */
export function deleteSegment(baseUrl: string, headers: Record<string, string>, segmentId: string): FilteredResponse { // delete segment
    const url = baseUrl + `/cxs/segments/${segmentId}`; // segment endpoint

    return callUnomi("DELETE", url, null, headers, 200); // return result object
}